# ufdbGuard.spec.CentOS7

%global _hardened_build 1
%global version 1.32.1

# no stripping of the binaries
%global __os_install_post      %{nil}
%define debug_package          %{nil}
%define __strip                /bin/true

### %__global_cflags        -O2 -g -pipe -Wall -Wp,-D_FORTIFY_SOURCE=2 -fexceptions -fstack-protector-strong --param=ssp-buffer-size=4 -grecord-gcc-switches %{_hardened_cflags} %{_performance_cflags}
%define __global_cflags     -O2 -g -pipe -Wall -grecord-gcc-switches %{_hardened_cflags} %{_performance_cflags}

Name: ufdbGuard
Version: %{version}
Release: CentOS7
Summary: ufdbGuard is a URL filter for Squid
License: GNU General Public License v2.0 only
Group: Internet/Proxy

# FHS says no package may have files under /usr/local nor /opt
# Prefix: /usr/local/ufdbguard
Prefix: /usr

Provides: ufdbguardd
Provides: ufdbgclient
Provides: ufdbhttpd
Provides: ufdbsignal
Provides: ufdb-pstack
Provides: ufdbpeek
Provides: ufdbGenTable, ufdbConvertDB
Provides: ufdbUpdate
Provides: ufdbAnalyse
Provides: ufdb_analyse_urls, ufdb_analyse_users, ufdb_top_urls, ufdb_top_users

URL: http://www.urlfilterdb.com/

# The sources for many versions of ufdbGuard are on sourceforge.net (Source0)
# The latest version can also be downloaded from URLfilterDB (Source1)
Source: www.urlfilterdb.com/files/downloads/%{name}-%{version}.tar.gz
# Source0: http://sourceforge.net/projects/ufdbguard/
# Source1: http://www.urlfilterdb.com/en/downloads/software_doc.html

# Buildroot: /local/src/ufdbGuard-%{version}
# Buildroot: .
BuildRoot:              %{_tmppath}/%{name}-%{version}-%{release}-buildroot-%(%{__id_u} -n)

# required packages for ufdbguardd
Requires: glibc >= 2.17
Requires: openssl >= 1.0.1e
Requires: bzip2-libs >= 1.0.6
Requires: zlib >= 1.2.7
# required packages for ufdbUpdate
Requires: wget >= 1.14
Requires: tar, gzip
# require dpackages for ufdb-pstack
Requires: gdb >= 7.6.1
Requires: yum-utils >= 1.1.31
# required packages for analysis scripts
Requires: perl
%global __requires_exclude perl\\(FCGI
# squid is required but may be installed from source and not using an RPM,
# or ufdbguard is used on a system where squid is not installed.
# Requires: squid
# gdb is highly recommended but not a requirement
# Requires: gdb

Buildrequires: openssl-devel >= 1.0.1e
Buildrequires: bzip2-devel >= 1.0.6
Buildrequires: zlib-devel >= 1.2.7
Buildrequires: make, gcc, bison, flex
Buildrequires: bind-utils

# TODO: %_initddir is macro for /etc/rc.d/init.d
Requires(post): chkconfig
Requires(preun): chkconfig
Requires(preun): initscripts
Requires(pre): shadow-utils

%description
ufdbGuard is a free URL filter for Squid with additional features like
SafeSearch enforcement for a large number of search engines, safer HTTPS 
visits and dynamic detection of proxies (URL filter circumventors).

ufdbGuard supports free and commercial URL databases that can be
downloaded from various sites and vendors.
You can also make your own URL database for ufdbGuard.


%post

echo >&2
echo "ufdbGuard is installed." >&2
echo "See the Reference Manual for further instructions and configuration." >&2
echo "Seek help at http://www.urlfilterdb.com in case you have a question or an issue." >&2
echo >&2

job=`grep ufdbUpdate /var/spool/cron/* 2>/dev/null | grep -v "^#" `
if [ "$job" = "" ]
then
   echo "There is not yet a cron job for ufdbUpdate  *****" >&2
   echo >&2
fi

# This adds the proper /etc/rc*.d links for the script
/sbin/chkconfig --add ufdb
#
echo "#!/usr/bin/sh" > /tmp/ufdb.postinstall
echo "echo Updating debuginfo ..." >> /tmp/ufdb.postinstall
echo "debuginfo-install -y -q glibc >/dev/null 2>&1" >> /tmp/ufdb.postinstall
echo "debuginfo-install -y -q zlib >/dev/null 2>&1" >> /tmp/ufdb.postinstall
echo "debuginfo-install -y -q bzip2 >/dev/null 2>&1" >> /tmp/ufdb.postinstall
echo "debuginfo-install -y -q openssl >/dev/null 2>&1" >> /tmp/ufdb.postinstall
chmod +x /tmp/ufdb.postinstall
#
echo "The installation of the ufdbGuard package is almost finished." >&2
echo "Execute /tmp/ufdb.postinstall to update debuginfo for glibc, zlib, bzip2 and openssl.  *****" >&2
echo >&2
#
# TODO: run check_dns

%preun
if [ $1 = 0 ] ; then
   /sbin/service ufdb stop >/dev/null 2>&1
   /sbin/chkconfig --del ufdb
fi


# for pre-F13:
%clean
[ %{buildroot} != "/" ] && echo rm -rf %{buildroot}

# ufdbGuard is installed with user ufdb and group ufdb
%pre
# set -x
getent group ufdb >/dev/null || groupadd -r ufdb
getent passwd ufdb >/dev/null || \
useradd -r -g ufdb -d /var/ufdbguard -M -s /sbin/nologin \
-c "ufdbGuard URL filter" ufdb
exit 0

%prep
# echo prep in %{buildroot}
# set -x
# TODO %setup -q
%setup -q

%build
echo build in `pwd`
%configure \
	--with-ufdb-user=ufdb \
	--prefix=/usr \
	--with-ufdb-bindir=/usr/sbin \
	--with-ufdb-piddir=/var/run/ufdbguard \
	--with-ufdb-mandir=/usr/share/man \
	--with-ufdb-images_dir=/var/ufdbguard/images \
	--with-ufdb-logdir=/var/ufdbguard/logs \
	--with-ufdb-samplesdir=/var/ufdbguard/samples \
	--with-ufdb-config=/etc/ufdbguard \
	--with-ufdb-dbhome=/var/ufdbguard/blacklists 

%{__make} %{?_smp_mflags}

%install
# echo install
# env
[ %{buildroot} != "/" ] &&   rm -rf %{buildroot}
%{__make} DESTDIR=%{buildroot} mkdirsredhatcentos install
# the install makes a backup of the conf file that we do not want in the package
rm -f %{buildroot}/etc/ufdbguard/ufdbGuard.conf.pre-v1.32

# echo
# echo "The configuration file of ufdbGuard is /etc/ufdbguard/ufdbGuard.conf"
# echo "The system configuration file for the ufdbGuard Software Suite is /etc/sysconfig/ufdbguard"

# ufdbsignal is suid-root since it must be able to send a signal to ufdbguardd.
# ufdbsignal is a very simple program which checks the uid to see if the user is permitted to send a signal.
# ufdbsignal reads the pid from /var/run/ufdbguardd/ufdbguardd.pid.

%verifyscript

if [ ! -f /etc/sysconfig/ufdbguard ]
then
   echo "/etc/sysconfig/ufdbguard does not exist." >&2
else
   eval `grep "^DOWNLOAD_USER=" /etc/sysconfig/ufdbguard`
   if [ "$DOWNLOAD_USER" = "" ]
   then
      echo "The username for periodical downloads of the URL database is not set." >&2
      echo "Edit /etc/sysconfig/ufdbguard and set DOWNLOAD_USER and DOWNLOAD_PASSWORD." >&2
   else
      echo "DOWNLOAD_USER is set to $DOWNLOAD_USER in /etc/sysconfig/ufdbguard"
   fi
fi

if [ ! -f /etc/ufdbguard/ufdbGuard.conf ]
then
   echo "/etc/ufdbguard/ufdbGuard.conf does not exist."
else
   set -- `grep ^dbhome /etc/ufdbguard/ufdbGuard.conf`
   # must get rid of quotes or else "if [ ! -d $DBDIR ]" fails :-(
   DBDIR=`echo ${2:-notset} | sed -e 's,",,g' `
   if [ $DBDIR = notset ]
   then
      DBDIR=/var/ufdbguard/blacklists
      echo "/etc/ufdbguard/ufdbGuard.conf: dbhome is not set" >&2
      echo "Using default value for dbhome: $DBDIR" >&2
   fi
   if [ ! -d $DBDIR ]
   then
      echo "/etc/ufdbguard/ufdbGuard.conf: dbhome $DBDIR: directory does not exist" >&2
   else
      if [ ! -d $DBDIR/adult  -o  ! -d $DBDIR/checked ]
      then
	 echo "/etc/ufdbguard/ufdbGuard.conf: dbhome $DBDIR:" >&2
         echo "The directory for the URL database does not contain subdirectories for adult and/or checked." >&2
	 echo "This means that the URL database of URLfilterDB is not used." >&2
	 echo "If you intend to use the URL database of URLfilterDB, make sure that " >&2
	 echo "\"ufdbUpdate [-v]\" runs without errors to download the URL database." >&2
	 echo "See the Reference Manual for more information." >&2
      fi
   fi
fi

exit 0


%postun

job=`grep ufdbUpdate /var/spool/cron/* 2>/dev/null | grep -v "^#" `
if [ "$job" != "" ]
then
   echo "Note: there is still a cron job for ufdbUpdate." >&2
fi

exit 0


# %config
# /etc/sysconfig/ufdbguard
# /etc/ufdbguard/ufdbGuard.conf

%files
%defattr(-,root,root,-)
/etc/init.d/ufdb
%config %attr(-,ufdb,ufdb) /etc/sysconfig/ufdbguard
%config %attr(-,ufdb,ufdb) /etc/ufdbguard/ufdbGuard.conf
/var/ufdbguard/images/default.flv
/var/ufdbguard/images/default.mp3
/var/ufdbguard/images/default.mpeg
/var/ufdbguard/images/default.wmv
/var/ufdbguard/images/forbidden-normal-de.png
/var/ufdbguard/images/forbidden-normal-en.png
/var/ufdbguard/images/forbidden-normal-es.png
/var/ufdbguard/images/forbidden-normal-fr.png
/var/ufdbguard/images/forbidden-normal-it.png
/var/ufdbguard/images/forbidden-normal-nl.png
/var/ufdbguard/images/forbidden-normal-pl.png
/var/ufdbguard/images/forbidden-normal-pt.png
/var/ufdbguard/images/forbidden-normal-sv.png
/var/ufdbguard/images/forbidden-normal-tr.png
/var/ufdbguard/images/no-ads.png
/var/ufdbguard/images/smallcross.png
/var/ufdbguard/images/square.png
/var/ufdbguard/images/transparent.png
/var/ufdbguard/samples/execdomainlist.sh
/var/ufdbguard/samples/execuserlist.sh
/var/ufdbguard/samples/URLblocked.cgi
/usr/sbin/ufdb-pstack
/usr/sbin/ufdbAnalyse
/usr/sbin/ufdbConvertDB
/usr/sbin/ufdbGenTable
/usr/sbin/ufdbUpdate
/usr/sbin/ufdb_analyse_urls
/usr/sbin/ufdb_analyse_users
/usr/sbin/ufdb_top_urls
/usr/sbin/ufdb_top_users
/usr/sbin/ufdbgclient
/usr/sbin/ufdbguardd
/usr/sbin/ufdbhttpd
%attr(4755,root,root) /usr/sbin/ufdbsignal
/usr/share/man/man1/ufdb_analyse_urls.1
/usr/share/man/man1/ufdb_analyse_users.1
/usr/share/man/man1/ufdb_top_urls.1
/usr/share/man/man1/ufdb_top_users.1
/usr/share/man/man1/ufdbAnalyse.1
/usr/share/man/man8/ufdbgclient.8
/usr/share/man/man8/ufdbguardd.8
/usr/share/man/man8/ufdbhttpd.8
/usr/share/man/man8/ufdbupdate.8
%dir %attr(-,ufdb,ufdb) /var/ufdbguard/blacklists
%dir %attr(-,ufdb,ufdb) /var/ufdbguard/blacklists/security
# The cacerts get updated by ufdbUpdate:
%verify(not md5 size mtime) %attr(644,ufdb,ufdb) /var/ufdbguard/blacklists/security/cacerts
# log files go to /var/ufdbguard/logs
%dir %attr(-,ufdb,ufdb) /var/ufdbguard/logs
# pid files go to /var/run/ufdbguard
%dir %attr(755,ufdb,ufdb) /var/run/ufdbguard

%doc README CHANGELOG
# TODO

%changelog
* Wed Oct 12 2016 Marcus Kool <marcus-dot-kool@urlfilterdb.com> - 1.32.1
Enhancement: New engine for HTTPS probing: do not probe known sites.
Enhancement: increase HTTPS queue size from 800 to 1500.
Enhancement: more content guessing in ufdbhttpd
Enhancement: Skype SSL certificate for pipe.skype.com is detected
Enhancement: ufdbgclient reconnects faster using last_failed_connect_time
Enhancement: use zlib compression: up to 5 times faster than the default bzip2 compression.
   This also implies that the URL database reload time is reduced.
Enhancement: ufdbGenTable uses a new internal data structure and is 7 times faster for large URL tables.
Enhancement: ufdbGuardd now issues errors in the log file when squid is configured to use helper concurrency but ufdbgclient does not have the -C option.
Enhancement: new pseudo categories: "localnet"
Enhancement: ufdbGenTable recognizes many new TLDs
Enhancement: ufdbUpdate can send email if it fails to download the URL database or fails to signal ufdbguardd.
Enhancement: URL categories can now have a mix of IP and usernames.
Enhancement: support ssl-bump with Squid 3.5.20 and higher
   Squid 3.5.x needs squid.conf with
   url_rewrite_extras "%>a/%>A %un %>rm bump_mode=%ssl::bump_mode sni=\"%ssl::>sni\" referer=\"%{Referer}>h\""
Enhancement: crash reports are uploaded to URLfilterDB for fast bug fixing.
   The option 'upload-crash-reports on|off' (default is on) controls if the crash report is uploaded.
Enhancement: ufdbGuard can match and block URLs with parameters based on the value of a parameter or the existence of a parameter
Enhancement: detect certificates of Anydesk.com and teamviewer (and new security options allow-teamviewer-over-https and allow-anydesk-over-https)
Enhancement: uncategorised URLs are now stored in a faster hash table and referers are also kept there.
Enhancement: ufdbhttpd has now support for POST
Enhancement: ufdbhttpd has support for social-badges category
Enhancement: new parameter: use-ipv6-on-wan (default: on) to enable IPv6 for connecting to all servers on the WAN (HTTPS probes and updates.urlfilterdb.com)
Enhancement: the time and date specifications are more flexible and now also accepts "8:00" where previously only "08:00" was accepted. same for leading zeroes in dates.
Enhancement: new parameter: pidfile
Enhancement: new parameter: num-worker-threads
Enhancement: new parameter: block-bumped-connect
Enhancement: new parameter: evaluate-and
Enhancement: new parameter: evaluate-or  (default)
Enhancement: Added checks for fatal errors: mix of exec and non-exec userlists, more than one execuserlist
Enhancement: Added check for error: category has no content definition
Enhancement: increased performance when a source contains many usernames
Enhancement: new security option https-prohibit-insecure-sslv3
Enhancement: new keyword "sender-email"
Enhancement: %C in the redirection URL shows all categories for the blocked URL.
Enhancement: the results of execuserlist(command) are cached to prevent delays in environments where it may take a long time to produce the userlist.
Fix: on some systems, when agressive HTTPS probing is used and the configuration is reloaded, the reload thread could not obtain a lock.
Fix: ufdbUpdate always had exit status 0
Fix: compilation failed on OSX
Fix: fix syntax error in execuserlist.sh
Fix: if the PIDDIR does not exist, ufdbguardd creates it, now also with the correct permissions.
Fix: add protection against oversized labels inside FQDN
Fix: emails by ufdbguardd have now a "Date:" header.
Fix: ufdbGenTable stripped 'www.' from 'www.com' which caused all .com domains to match.
Fix: count number of active workers and generate errors if this is higer than number of workers
Fix: ufdbpeek crashed when logging very large messages (buffer overflow in ufdbLog)
Fix: fix a few small memory leaks
Fix: ufdbGenTable may produce corrupted URL tables is a domain contains '..'
Fix: category with "option safesearch" blocked access
Change: removed -N option since this function is controlled by a configuration parameter
Change: uncategorised URLs are now uploaded with HTTPS instead of HTTP.

