/* 
 * strcmpurlpart.static.c - URLfilterDB
 *
 * ufdbGuard is copyrighted (C) 2005-2016 by URLfilterDB with all rights reserved.
 *
 * Parts of the ufdbGuard daemon are based on squidGuard.
 * This module is NOT based on squidGuard.
 *
 * $Id: strcmpurlpart.static.c,v 1.18 2016/09/06 00:49:54 root Exp root $
 */

/* =================================================================================
 * NOTE: with more recent versions of glibc, strcmp uses SSE4.2 instructions and is 
 * ultrafast but still slower than our loop - the SSE4.2 version is only faster for longer strings.
 */


#if (__GNUC__ > 4)  ||  (__GNUC__ == 4  &&  __GNUC_MINOR__ >= 4)
#pragma GCC push_options
#pragma GCC optimize ("O3")
#endif

/* must use gcc -msse4.2 in order to get __SSE4_2__ defined : */

/*
#ifdef __SSE4_2__
#include "inline-strcmp-sse42.c"
#endif
*/

#define UFDB_USE_LARGE_LOOP_UNROLL 1


#if __SSE4_2__ && 0

#include <x86intrin.h>

UFDB_GCC_INLINE UFDB_GCC_HOT 
static int __ufdb_inline_strcmp_sse42( const char * url, const char * table )
{
   __m128i regurl;
   __m128i regti;
   int idx;

   while (1)
   {
      regurl = _mm_loadu_si128( (__m128i const *) url );
      regti  = _mm_loadu_si128( (__m128i const *) table );
      idx = _mm_cmpistri( regurl, regti, _SIDD_UBYTE_OPS | _SIDD_CMP_EQUAL_EACH | _SIDD_MASKED_NEGATIVE_POLARITY );
      if (idx < 16)
	 return (url[idx] - table[idx]);
      url += 16;
      table += 16;
   }
   return 0;
}
#endif


UFDB_GCC_INLINE UFDB_GCC_HOT 
static int strcmpURLpart( const char * url, const char * table )
{
   if (*table != '/')
   {
#if __SSE4_2__ && 0
      /* strcmp() is never inlined by gcc since there is no constant string.
       * strcmp() in glibc 2.12 uses sse[234] if available, and is considered fast, but
       * still is slower than the loop below (GCC 4.4.7 and GLIBC 2.12).
       * So lets try it with our inlined SSE4.2 strcmp() :
       * The inlined SSE 4.2 strcmp() is slower than the byte-by-byte comparison BECAUSE
       * most strings that are compared differ in the first few bytes and with the simpler
       * byte-by-byte comparison loop the final result is computed faster.
       *
       * In a nutshell: strcmp() and SSE 4.2 are useful/faster for larger strings but not for
       * the set of small strings that we deal with here.
       */
      /* return __ufdb_strcmp_sse42( url, table ); */		/* using assembly */
      return __ufdb_inline_strcmp_sse42( url, table );		/* using compiler intrinsics */
#else

#if 0
      /* inlined: return strcmp( url, table ); */
      while (*url != '\0'  &&  (*url - *table) == 0)
      {
	 url++;
	 table++;
      }
      return *url - *table;
#else

      do
      {
	 char d;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

#ifdef UFDB_USE_LARGE_LOOP_UNROLL
	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;

	 d = *url - *table;
	 if (d != 0  ||  *url == '\0')
	    return d;
	 url++;
	 table++;
#endif
      } while (1);
#endif

#endif
   }
   else
   {
      while (*table != '\0')
      {
	 if ((*url - *table) != 0)
	 {
	    if (*table == '|' && *url == '\0')
	       return 0;
	    else
	       return *url - *table;
	 }
	 url++;
	 table++;
      }
   }

   return 0;
}

#if (__GNUC__ > 4)  ||  (__GNUC__ == 4  &&  __GNUC_MINOR__ >= 4)
#pragma GCC pop_options
#endif

